/**
 * (c) Atos IT Solutions and Services GmbH, 2011.
 * Licensed under EUPL 1.1
 */

package com.sag.osami.api.context;

import java.util.Date;

/**
 * Instances of the ContextEvent class represent a single context event and its constituents, since a context is a container for the context data + context
 * management data (ID, timestamp, origin, type, origin context store).
 * 
 * <ul>
 * <li>The <em>ID</em> is a unique number which identifies a context event within the context store
 * <li>The <em>timestamp</em> is stored as a standard {@link Date} object which allows a resolution of milliseconds
 * <li>The <em>origin</em> of the context event is identified by a unique ID that should allow the programmer to unambiguously determine which component or
 * technical sub-system has created the event
 * <li>The <em>type</em> is an identifier for the data object which is the payload of the context event (however, the type doesn't have to be the fully
 * qualified class name)
 * <li>The <em>data</em> is an arbitrary object which represents the payload of the context event.
 * <li>The <em>origin context store</em> of the context event reveals in which context store this event was originally created (introduced for future releases
 * which may allow forwarding of context events between different context stores; see also {@link IContextStore#getStoreID()})
 * </ul>
 * 
 * This class is immutable: once a context event was produced, there is no possibility to corrupt its data (e.g. within a handler).
 */
public final class ContextEvent {

	private final Long id;
	private final Object data;
	private final String originID;
	private final String contextStoreOriginID;
	private final Date timestamp;
	private final String type;

	/**
	 * Constructor (mainly) for public use, taking the parameters as the values for the context event, but {@code null} as {@code ID} and
	 * {@code contextStoreOriginID} (both determined internally by context store)
	 * 
	 * @param data
	 *            payload data of the context event, may be null
	 * @param originID
	 *            origin ID of the context event
	 * @param timestamp
	 *            creation date of the context event
	 * @param type
	 *            type of the context event
	 */
	public ContextEvent(Object data, String originID, Date timestamp, String type) {
		this(null, data, originID, timestamp, type, null);
	}

	/**
	 * Constructor (mainly) for internal use, using the parameters as the values for the context event.
	 * 
	 * @param id
	 *            unique numeric ID of the context event
	 * @param data
	 *            payload data of the context event, may be null
	 * @param originID
	 *            origin ID of the context event
	 * @param timestamp
	 *            creation date of the context event
	 * @param type
	 *            type of the context event
	 * @param contextStoreOriginID
	 *            name of the context store where the context event was originally created
	 */
	public ContextEvent(Long id, Object data, String originID, Date timestamp, String type, String contextStoreOriginID) {
		if (originID == null)
			throw new NullPointerException("originID must not be null");
		if (timestamp == null)
			throw new NullPointerException("timestamp must not be null");
		if (type == null)
			throw new NullPointerException("type must not be null");

		this.id = id;
		this.data = data;
		this.contextStoreOriginID = contextStoreOriginID;
		this.originID = originID;
		this.timestamp = timestamp;
		this.type = type;
	}

	/**
	 * Gets the numeric, unique ID of the context.
	 * 
	 * @return the numeric, unique ID of the context
	 */
	public Long getID() {
		return id;
	}

	/**
	 * Gets the type of the context. It is represented by a plain string, which meaning should be unique for the application. If possible, standardized class
	 * names/ontological information should be used.
	 * <p>
	 * Example: <code>"TEMP_C"</code> (for temperature in Celsius)
	 * 
	 * @return the type of the context
	 */
	public String getType() {
		return type;
	}

	/**
	 * Gets the payload data of the context.
	 * 
	 * @return the payload data of the context (probably needs type cast)
	 */
	public Object getData() {
		return data;
	}

	/**
	 * Gets the origin of the context. The origin ID should identify the component or sub-system where the context was created (may denote hardware or
	 * software).
	 * <p>
	 * Example: <code>"COM/SAG/OSAMI/SENSORS/VITAL"</code> for some sensor system
	 * 
	 * @return the origin of the context
	 */
	public String getOriginID() {
		return originID;
	}

	/**
	 * Gets the time of creation of the context.
	 * 
	 * @return the time of "creation" of the context
	 */
	public Date getTimestamp() {
		return timestamp;
	}

	/**
	 * Gets the origin ContextStore of the context, i.e. the (unique) name of the ContextStore where the context was created originally.
	 * 
	 * @return the origin ContextStore of the context
	 */
	public String getContextStoreOriginID() {
		return contextStoreOriginID;
	}
}
