/**
 * (c) Atos IT Solutions and Services GmbH, 2011.
 * Licensed under EUPL 1.1
 */
package com.sag.osami.api.context;

import java.util.List;

/**
 * This interface provides low-level methods for storing, loading, deleting and querying context events in the ContextStore. No method (in particular
 * {@link #store(ContextEvent)}) will ever invoke any {@link IContextHandler} logic, nor will it register new context types in the ContextStore etc. All methods
 * strictly handle storage concerns and nothing further.
 * <p>
 * <b>Important:</b> This interface is intended for public use. It is NOT the actual interface which is used internally for storage management. If you plan to
 * implement a custom storage management class, see {@link com.sag.osami.contextstore.serialization.ISerializedContextStorage}.
 * <p>
 * <i>OSGi notice:</i> In OSGi environments, the storage implementation should be registered as a service providing this interface. When no custom storage
 * back-end is provided, the shipped one is registered accordingly (see {@link com.sag.osami.contextstore.db.DerbyStorage}).
 */
public interface IContextStorage {

	/**
	 * Loads a context event with the given ID.
	 * 
	 * @param contextId
	 *            ID of the context event
	 * @return loaded context event (or null, if there is no context event with given ID)
	 */
	ContextEvent load(Long contextId);

	/**
	 * Loads several context events with the given IDs. The order within the result is unspecified.
	 * 
	 * @param contextIds
	 *            IDs of the context events.
	 * @return loaded context events (or a zero-length array, if there are no context events with given IDs)
	 */
	ContextEvent[] load(List<Long> contextIds);

	/**
	 * Lazy variant of {@link #load(List)}, i.e. it loads one ContextEvent at a time during iteration. This operation is optional.
	 * 
	 * @param contextIds
	 *            IDs of the context events.
	 * @throws UnsupportedOperationException
	 *             if the underlying storage implementation does not support lazy querying
	 * @see #load(List)
	 */
	Iterable<ContextEvent> loadLazy(List<Long> contextIds) throws UnsupportedOperationException;

	/**
	 * Queries several context events.
	 * 
	 * @param query
	 *            Its language/format depends on the actual storage management class.
	 * @param params
	 *            Should be used for parameter binding à la JDBC:
	 *            {@code query("SELECT * FROM contextstore WHERE type = ? AND timestamp >= ?", "someType", someDate.getTime());}
	 * @return queried context events (or a zero-length array, if there are no according context events)
	 */
	ContextEvent[] query(String query, Object... params);

	/**
	 * Lazy variant of {@link #query(String, Object...)}, i.e. it loads one ContextEvent at a time during iteration. This operation is optional.
	 * 
	 * @param query
	 *            The context event query.
	 * @param params
	 *            Parameters for the query.
	 * @throws UnsupportedOperationException
	 *             if the underlying storage implementation does not support lazy querying
	 * @see #query(String, Object...)
	 */
	Iterable<ContextEvent> queryLazy(String query, Object... params) throws UnsupportedOperationException;

	/**
	 * Stores a context event object persistently.
	 * 
	 * @param context
	 *            context event object to be stored.
	 */
	void store(ContextEvent context);

	/**
	 * Deletes a context event with given ID.
	 * 
	 * @param contextId
	 *            ID of context event to be deleted.
	 */
	void delete(Long contextId);

	/**
	 * Deletes several context events with given IDs.
	 * 
	 * @param contextIds
	 *            IDs of context events to be deleted.
	 */
	void delete(List<Long> contextIds);

	/**
	 * Deletes several context events defined by the query.
	 * 
	 * @param query
	 *            Defines the context events to be deleted.
	 * @param params
	 *            Parameters for the query.
	 * @see #query(String, Object...)
	 */
	void delete(String query, Object... params);
}
