/**
 * (c) Atos IT Solutions and Services GmbH, 2011.
 * Licensed under EUPL 1.1
 */
package com.sag.osami.api.context;

import java.util.Date;

/**
 * The ContextStore manages a persistent store for context events and manages context providers and consumers through a uniform interface. Context events are
 * dispatched to registered handlers according to the registered types of the handlers. Handlers can change the registration for context types during their
 * lifetime.
 * <p>
 * Since version 1.1.0, the semantics of the <code>newContextEvent</code> methods changed substantially: the events created this way are no longer processed
 * synchronously. Handlers execute their handling code in own threads.
 * <p>
 * Also since version 1.1.0, ContextStores now feature a (ideally unique) ID. This ID might be used in future releases, where communication between
 * ContextStores (e.g. on different hosts) might be possible.
 * <p>
 * <i>OSGi notice:</i> In OSGi environments, the ContextStore is registered as a service providing this interface.
 * 
 * @see IContextHandler
 */
public interface IContextStore extends IContextQuery {

	public final static String CONTEXT_STORE_ID = "CONTEXT_STORE_SYSTEM";

	public final static String TYPE_REGISTERED = "ContextStore/Type/Registered";
	public final static String TYPE_DEREGISTERED = "ContextStore/Type/Deregistered";

	/** This constant is used as a context type when a handler should receive all context events */
	public final static String TYPE_ALL = "*";

	/**
	 * Insert a new context with the given parameters in the ContextStore.
	 * 
	 * @param originID
	 *            the origin id of the new context event
	 * @param type
	 *            the type of the new context event
	 * @param data
	 *            the data of the new context event
	 * @param date
	 *            the date of the new context event
	 */
	public void newContextEvent(String originID, String type, Object data, Date date);

	/**
	 * Insert a new context with the given parameters in the ContextStore.
	 * 
	 * @param originID
	 *            the origin id of the new context event
	 * @param type
	 *            the type of the new context event
	 * @param data
	 *            the data of the new context event
	 * @param date
	 *            the date of the new context event
	 * @param persist
	 *            indicates whether context event should be persisted or not
	 */
	public void newContextEvent(String originID, String type, Object data, Date date, boolean persist);

	/**
	 * Insert a new context in the ContextStore.
	 * <p>
	 * <em>Note that the actual context event object won't be stored as-is in the ContextStore but instead is re-packaged in
	 * a new ContextEvent data object. Also, the properties ID and contextStoreOriginID stored in the given object will be
	 * ignored and determined internally. This method exists for convenience only.</em>
	 * 
	 * @param ctxEvent
	 *            the context event to be inserted
	 */
	public void newContextEvent(ContextEvent ctxEvent);

	/**
	 * Insert a new context in the ContextStore.
	 * <p>
	 * <em>Note that the actual context event object won't be stored as-is in the ContextStore but instead is re-packaged in
	 * a new ContextEvent data object. Also, the properties ID and contextStoreOriginID stored in the given object will be
	 * ignored and determined internally. This method exists for convenience only.</em>
	 * 
	 * @param ctxEvent
	 *            the context event to be inserted
	 * @param persist
	 *            indicates whether context event should be persisted or not
	 */
	public void newContextEvent(ContextEvent ctxEvent, boolean persist);

	/**
	 * Register a context handler for notification of new contexts (all types).
	 * 
	 * @param ctxHandler
	 *            the context handler
	 */
	public void registerHandler(IContextHandler ctxHandler);

	/**
	 * Register a context handler for notification of new contexts with any type in the given list.
	 * 
	 * @param ctxHandler
	 *            the context handler
	 * @param ctxTypes
	 *            the context type list
	 */
	public void registerHandler(IContextHandler ctxHandler, String[] ctxTypes);

	/**
	 * Deregister context handler.
	 * 
	 * @param ctxHandler
	 *            the context handler
	 */
	public void deregisterHandler(IContextHandler ctxHandler);

	/**
	 * Activate notification. If handler had notification deactivated, now events are being delivered again.
	 * <p>
	 * Has no effect when notification was already active.
	 * 
	 * @param ctxHandler
	 *            the context handler
	 */
	public void activateEventsForHandler(IContextHandler ctxHandler);

	/**
	 * Deactivate notification. ContextStore no longer dispatches newContextEvents to the handler.
	 * <p>
	 * This is useful if the handler does some time-consuming operations while processing, or needs to be temporarily disabled. Despite the deactivation, the
	 * normal operation can be resumed without having to register again.
	 * <p>
	 * Has no effect when notification was already inactive.
	 * 
	 * @param ctxHandler
	 *            the context handler
	 */
	public void deactivateEventsForHandler(IContextHandler ctxHandler);

	/**
	 * Register context type in the ContextStore.
	 * <p>
	 * This routine is useful, if a context source is not available in the beginning, but there are already handlers registering for future context events of
	 * the given type.
	 * 
	 * @param ctxType
	 *            the context type
	 */
	public void registerContextType(String ctxType);

	/**
	 * Deregister context type in the ContextStore.
	 * 
	 * @param ctxType
	 *            the context type
	 */
	public void deregisterContextType(String ctxType);

	/**
	 * Checks if given context type is registered in the ContextStore.
	 * 
	 * @param ctxType
	 *            the context type to be checked
	 * 
	 * @return boolean whether type is registered in the context store
	 */
	public boolean isRegisteredContextType(String myType);

	/**
	 * Returns all registered context types.
	 * 
	 * @return list of all registered context types
	 */
	public String[] getRegisteredContextTypes();

	/**
	 * Returns all registered context types for committed handler.
	 * 
	 * @return list of all registered context types
	 */
	public String[] getHandlersRegisteredContextTypes(IContextHandler handler);

	/**
	 * Register the context handler for the given context type.
	 * 
	 * @param ctxHandler
	 *            the context handler
	 * @param ctxType
	 *            the context type
	 */
	public void registerHandlerForContextType(IContextHandler ctxHandler, String ctxType);

	/**
	 * Deregister the context handler for the given context type.
	 * 
	 * @param ctxHandler
	 *            the context handler
	 * @param ctxType
	 *            the context type
	 */
	public void deregisterHandlerForContextType(IContextHandler ctxHandler, String ctxType);

	/**
	 * Returns the ID of the ContextStore.
	 * 
	 * @return context store ID
	 */
	public String getStoreID();

	/**
	 * Returns the storage handler of the context store, which handles storing, loading, querying and deleting context events
	 * 
	 * @return context storage handler
	 */
	public IContextStorage getStorage();
}