/**
 * (c) Atos IT Solutions and Services GmbH, 2011.
 * Licensed under EUPL 1.1
 */
package com.sag.osami.contextstore.osgi;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.osgi.util.tracker.ServiceTracker;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.sag.osami.api.context.IContextHandler;
import com.sag.osami.api.context.IContextQuery;
import com.sag.osami.api.context.IContextStorage;
import com.sag.osami.api.context.IContextStore;
import com.sag.osami.api.context.IContextStoreID;
import com.sag.osami.contextstore.ContextStore;
import com.sag.osami.contextstore.ContextStoreBuilder;
import com.sag.osami.contextstore.serialization.ISerializationHandler;
import com.sag.osami.contextstore.serialization.ISerializedContextHandler;
import com.sag.osami.contextstore.serialization.ISerializedContextProducer;
import com.sag.osami.contextstore.serialization.ISerializedContextReceiver;
import com.sag.osami.contextstore.storage.ISerializedContextStorage;
import com.sag.osami.contextstore.storage.jdbc.IJDBCDialect;

/**
 * ContextStore OSGi startup and cleanup class. It searches for the following ContextStore components via the OSGi service registry:
 * <p>
 * <ul>
 * <li>{@link IContextStoreID}
 * <li>{@link ISerializationHandler}
 * <li>{@link IJDBCDialect}
 * <li>{@link ISerializedContextStorage}
 * </ul>
 * 
 * It falls back to the defaults of the {@link ContextStoreBuilder}, if a component is not found in the service registry, and in turn registers them as a
 * service.
 * <p>
 * It also enables service trackers which are responsible for automatically register {@link IContextHandler}s (and {@link ISerializedContextHandler}s,
 * {@link ISerializedContextProducer}s) if they are registered as OSGi services.
 */
public class OsgiContextStoreActivator implements BundleActivator {

	private ServiceTracker handlerTracker, serializedTracker, serializedProducerTracker;

	private List<ServiceRegistration> svcRegs = new LinkedList<ServiceRegistration>();

	// these flags indicate whether the storage and serializer should be registered in the OSGi service registry (i.e. the default implementations were
	// picked <-> no according services were registered yet)
	private boolean omitSerializerReg, omitStorageReg;

	private Logger log = LoggerFactory.getLogger(getClass());

	private static BundleContext context;

	public static BundleContext getContext() {
		return context;
	}

	public void start(BundleContext context) throws Exception {
		OsgiContextStoreActivator.context = context;

		ContextStoreBuilder ctxBuild = new ContextStoreBuilder();

		// store ID service lookup
		IContextStoreID id = getService(IContextStoreID.class);
		if (id != null)
			ctxBuild.storeID(id.storeID());

		// serializer service lookup
		ISerializationHandler serializer = getService(ISerializationHandler.class);
		if (serializer != null) {
			omitSerializerReg = true;
			ctxBuild.serializer(serializer);
		}

		// JDBC dialect service lookup
		IJDBCDialect jdbcDialect = getService(IJDBCDialect.class);
		if (jdbcDialect != null) {
			ctxBuild.jdbcDialect(jdbcDialect);
		}

		// storage service lookup
		ISerializedContextStorage storage = getService(ISerializedContextStorage.class);
		if (storage != null) {
			omitStorageReg = true;
			ctxBuild.storage(storage);
		}

		ContextStore store = ctxBuild.build();

		// default storage registration
		if (!omitStorageReg) {
			String[] regClasses = { ISerializedContextStorage.class.getName(), IContextStorage.class.getName() };
			svcRegs.add(context.registerService(regClasses, store.getStorage(), null));
		}

		// default serializer registration
		if (!omitSerializerReg) {
			svcRegs.add(context.registerService(ISerializationHandler.class.getName(), store.getSerializationHandler(), null));
		}

		OsgiContextStoreSetter customizer = new OsgiContextStoreSetter(store);
		handlerTracker = new ServiceTracker(context, IContextHandler.class.getName(), customizer);
		handlerTracker.open(true);
		serializedTracker = new ServiceTracker(context, ISerializedContextHandler.class.getName(), customizer);
		serializedTracker.open(true);
		serializedProducerTracker = new ServiceTracker(context, ISerializedContextProducer.class.getName(), customizer);
		serializedProducerTracker.open(true);

		// ContextStore registration
		String[] regClasses = { IContextStore.class.getName(), ISerializedContextReceiver.class.getName(), IContextQuery.class.getName() };
		svcRegs.add(context.registerService(regClasses, store, null));

		log.info("ContextStore initialized");
	}

	public void stop(BundleContext context) throws Exception {
		handlerTracker.close();
		serializedTracker.close();
		serializedProducerTracker.close();

		Collections.reverse(svcRegs);
		for (ServiceRegistration reg : svcRegs)
			if (reg != null)
				reg.unregister();
		svcRegs = null;

		OsgiContextStoreActivator.context = null;
	}

	@SuppressWarnings("unchecked")
	private static synchronized <T> T getService(Class<T> clazz) {
		ServiceReference svcRef = context.getServiceReference(clazz.getName());
		if (svcRef != null) {
			return (T) context.getService(svcRef);
		}
		return null;
	}
}