/**
 * (c) Atos IT Solutions and Services GmbH, 2011.
 * Licensed under EUPL 1.1
 */
package com.sag.osami.contextstore.storage.impl;

import java.io.StringReader;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.sag.osami.api.context.ContextEvent;
import com.sag.osami.contextstore.ContextStoreBuilder;
import com.sag.osami.contextstore.storage.DBConstants;
import com.sag.osami.contextstore.storage.SerializedContextStorage;
import com.sag.osami.contextstore.storage.impl.IteratorUtil.RowIterator;
import com.sag.osami.contextstore.storage.jdbc.DerbyDialect;
import com.sag.osami.contextstore.storage.jdbc.IJDBCDialect;
import com.sag.osami.contextstore.util.NamedThreadFactory;

/**
 * Context storage implementation based on JDBC. It is database-agnostic by using a {@link IJDBCDialect} interface which provides all database-specific SQL
 * statements. By default, {@link DerbyDialect} is used.
 * <p>
 * It is configurable using the {@link ContextStoreBuilder}. Following properties can be configured:
 * <p>
 * <ul>
 * <li>{@value #CONFIG_URL} : JDBC URL of the database
 * <li>{@value #CONFIG_AUTODROP} : flag which indicates whether an existing context store table should be auto-dropped on start
 * <li>{@value #CONFIG_USERNAME} : username for the database connection
 * <li>{@value #CONFIG_PASSWORD} : password for the database connection
 * </ul>
 * If a JDBC URL is configured, this class will try to derive the used dialect from this URL. For the JDBC URL <code>jdbc:mydb://...</code>, this dialect class
 * will be used:
 * 
 * <pre>
 * <code>com.sag.osami.contextstore.storage.jdbc.MydbDialect</code>
 * </pre>
 * 
 * The table layout is according to {@link DBConstants}.
 * <p>
 * An example query on the context store table, when using Derby, would look like this (example fetches last context event with given origin and type):
 * 
 * <pre>
 * public static ContextEvent queryLast(String type, String originID, IContextStore contextStore) {
 * 	String queryGetLastItem = &quot;SELECT * FROM contextstore WHERE type = ? AND originID = ? ORDER BY timestamp FETCH ONLY LAST ROW&quot;;
 * 	IContextEvent[] tmp = contextStore.query(queryGetLastItem, type, originID);
 * 	if (tmp.length != 0)
 * 		return tmp[0];
 * 	else
 * 		return null;
 * }
 * </pre>
 * <p>
 * Please note, that all column names (see table layout in {@link DBConstants}) of the ContextStore table must be in the result. While it is most convenient to
 * use <code>"SELECT * FROM contextstore ..."</code> to ensure this, it is not mandatory. The only requirement is, that the result contains at least all column
 * names, therefore allowing complex queries (JOINs etc.) to be performed.
 * <p>
 * <i>OSGi notice:</i> To conveniently use other dialects than Derby in an OSGi environment, deploy the according fragment bundle (MySQL, PostgreSQL and SQLite
 * are available) and further deploy a configuration file in another fragment bundle (with the ContextStore bundle as its host). Another way is to register your
 * own {@link IJDBCDialect} implementation as an OSGi service before the ContextStore initialization.
 * 
 * @see IJDBCDialect
 * @see ContextStoreBuilder
 * @see DBConstants
 */
public final class JDBCStorage extends SerializedContextStorage implements DBConstants {

	public static final String CONFIG_URL = "storage.jdbc.url";
	public static final String CONFIG_AUTODROP = "storage.jdbc.autodrop";
	public static final String CONFIG_USERNAME = "storage.jdbc.username";
	public static final String CONFIG_PASSWORD = "storage.jdbc.password";

	private static final String DIALECTS_PACKAGE = IJDBCDialect.class.getPackage().getName();

	private final Logger log = LoggerFactory.getLogger(JDBCStorage.class);

	private long lastID = 0;

	private final IJDBCDialect dialect;

	private final Set<Long> deletedIds;
	private final Connection conn;
	private final ExecutorService dbWorker;
	private final boolean autodrop;

	private static IJDBCDialect getDialectIntanceForUrl(String jdbcUrl) {
		try {
			String dialectClass = jdbcUrl.split(":")[1].toLowerCase();
			dialectClass = DIALECTS_PACKAGE + "." + Character.toUpperCase(dialectClass.charAt(0)) + dialectClass.substring(1) + "Dialect";
			System.err.println(dialectClass);
			return (IJDBCDialect) Class.forName(dialectClass).newInstance();
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Creates a JDBC-based storage with the given dialect, taking the JDBC URL etc. from the given properties
	 * 
	 * @param dialect
	 *            the JDBC dialect to be used
	 * @param jdbcProps
	 *            the configuration properties
	 */
	public JDBCStorage(IJDBCDialect dialect, Properties jdbcProps) {
		this(dialect, jdbcProps.getProperty(CONFIG_URL), jdbcProps.getProperty(CONFIG_USERNAME), jdbcProps.getProperty(CONFIG_PASSWORD), Boolean
				.parseBoolean(jdbcProps.getProperty(CONFIG_AUTODROP)));
	}

	/**
	 * Creates a JDBC-based storage, taking the JDBC URL etc. from the given properties, auto-detecting the dialect based on the JDBC URL
	 * 
	 * @param jdbcProps
	 *            the configuration properties
	 */
	public JDBCStorage(Properties jdbcProps) {
		this(getDialectIntanceForUrl(jdbcProps.getProperty(CONFIG_URL)), jdbcProps);
	}

	/**
	 * Creates a JDBC-based storage, auto-detecting the dialect based on the JDBC URL
	 * 
	 * @param jdbcUrl
	 *            the JDBC URL
	 * @param autodrop
	 *            whether the context store table should be auto-dropped
	 */
	public JDBCStorage(String jdbcUrl, boolean autodrop) {
		this(getDialectIntanceForUrl(jdbcUrl), jdbcUrl, null, null, autodrop);
	}

	/**
	 * Creates a JDBC-based storage, auto-detecting the dialect based on the JDBC URL
	 * 
	 * @param jdbcUrl
	 *            the JDBC URL
	 * @param jdbcUsername
	 *            the JDBC username
	 * @param jdbcPassword
	 *            the JDBC password
	 * @param autodrop
	 *            whether the context store table should be auto-dropped
	 */
	public JDBCStorage(String jdbcUrl, String jdbcUsername, String jdbcPassword, boolean autodrop) {
		this(getDialectIntanceForUrl(jdbcUrl), jdbcUrl, jdbcUsername, jdbcPassword, autodrop);
	}

	/**
	 * Creates a JDBC-based storage with the given dialect
	 * 
	 * @param dialect
	 *            the JDBC dialect to be used
	 * @param jdbcUrl
	 *            the JDBC URL
	 * @param autodrop
	 *            whether the context store table should be auto-dropped
	 */
	public JDBCStorage(IJDBCDialect dialect, String jdbcUrl, boolean autodrop) {
		this(dialect, jdbcUrl, null, null, autodrop);
	}

	/**
	 * Creates a JDBC-based storage with the given dialect
	 * 
	 * @param dialect
	 *            the JDBC dialect to be used
	 * @param jdbcUrl
	 *            the JDBC URL
	 * @param jdbcUsername
	 *            the JDBC username
	 * @param jdbcPassword
	 *            the JDBC password
	 * @param autodrop
	 *            whether the context store table should be auto-dropped
	 */
	public JDBCStorage(IJDBCDialect dialect, String jdbcUrl, String jdbcUsername, String jdbcPassword, boolean autodrop) {
		try {
			this.dialect = dialect;
			this.autodrop = autodrop;
			log.info("Initializing JDBC storage for ContextEvents [dialect: {}, url: {}, autodrop: {}", new Object[] { dialect.getClass().getName(), jdbcUrl,
					autodrop });

			Class.forName(dialect.driver());
			if (jdbcUsername == null)
				conn = DriverManager.getConnection(jdbcUrl);
			else
				conn = DriverManager.getConnection(jdbcUrl, jdbcUsername, jdbcPassword);

			deletedIds = Collections.synchronizedSet(new TreeSet<Long>());
			initTable();
			dbWorker = Executors.newSingleThreadScheduledExecutor(new NamedThreadFactory("JDBCStorage", false));
		} catch (Throwable t) {
			throw new RuntimeException("JDBCStorage as storage for ContextEvents could not be initialized", t);
		}
	}

	/**
	 * This method checks whether the ContextStore table already exists or not, by executing a SELECT query on the table. This is somewhat unclean, however, it
	 * is quite independent of the actually used JDBC driver.
	 * 
	 * @return whether the ContextStore table exists (true) or not (false)
	 * @throws SQLException
	 */
	private boolean tableExists() throws SQLException {
		Statement stmt = null;
		try {
			stmt = conn.createStatement();
			stmt.execute(dialect.tableExists());
			return true;
		} catch (SQLException e) {
			return false;
		} finally {
			if (stmt != null)
				stmt.close();
		}
	}

	private void initTable() throws SQLException {
		boolean tableExists = tableExists();

		if (autodrop && tableExists)
			dropTable();

		if (autodrop || !tableExists)
			createTable();

		if (!autodrop && tableExists)
			queryLastID();
	}

	private void dropTable() {
		Statement stmt = null;
		try {
			try {
				stmt = conn.createStatement();
				String dropSql = dialect.dropTable();
				log.info("Autop-dropping ContextStore table: " + dropSql);
				stmt.execute(dropSql);
			} finally {
				if (stmt != null)
					stmt.close();
			}
		} catch (SQLException e) {
			log.error("ContextStore table was not dropped!", e);
		}
	}

	private void createTable() throws SQLException {
		Statement stmt = null;

		try {
			stmt = conn.createStatement();
			String createSql = dialect.createTable();
			log.info("Creating ContextStore table: " + createSql);
			stmt.execute(createSql);
		} finally {
			if (stmt != null)
				stmt.close();
		}
	}

	@Override
	public void store(final boolean serialize, final ContextEvent context) {
		dbWorker.execute(new Runnable() {
			public void run() {
				try {
					if (deletedIds.remove(context.getID())) // context event with given ID was
															// already marked as to-be-deleted
															// before it had the chance to be
															// actually stored
						return;

					String[] serializedData = (serialize) ? serializer.serialize(context.getData()) : (String[]) context.getData(); // might be null
					PreparedStatement insertStmt = conn.prepareStatement(dialect.insertContextEvent());

					try {
						insertStmt.setLong(1, context.getID());
						insertStmt.setString(2, context.getOriginID());
						insertStmt.setString(3, context.getType());
						insertStmt.setString(6, context.getContextStoreOriginID());
						insertStmt.setTimestamp(7, new Timestamp(context.getTimestamp().getTime()));

						if (serializedData != null) {
							insertStmt.setString(4, serializedData[0]);
							insertStmt.setCharacterStream(5, new StringReader(serializedData[1]), serializedData[1].length());
						} else {
							insertStmt.setString(4, null);
							insertStmt.setString(5, null);
						}

						insertStmt.executeUpdate();
					} finally {
						if (insertStmt != null)
							insertStmt.close();
					}
				} catch (SQLException e) {
					log.error("Persisting of ContextEvent failed!", e);
				}
			}
		});
	}

	@Override
	public void delete(Long contextId) {
		try {
			PreparedStatement stmt = conn.prepareStatement(dialect.deleteContextEvent());
			stmt.setLong(1, contextId);
			// if context wasn't stored yet, mark as to-be-deleted
			if (stmt.executeUpdate() <= 0)
				deletedIds.add(contextId);
		} catch (SQLException e) {
			log.error("ContextEvent deletion request failed", e);
		}
	}

	@Override
	public void delete(List<Long> contextIds) {
		for (Long contextId : contextIds)
			delete(contextId);
	}

	@Override
	public void delete(String query, Object... params) {
		if (!query.toLowerCase().startsWith("delete from"))
			throw new IllegalArgumentException("A deletion-query must be actually a DELETE FROM SQL query.");
		try {
			parametrizedStatement(query, params).execute();
		} catch (SQLException e) {
			log.error("ContextEvent deletion request failed", e);
		}
	}

	@Override
	public Iterable<ContextEvent> queryLazy(final boolean deserialize, final String query, final Object... params) {
		return new Iterable<ContextEvent>() {
			@Override
			public Iterator<ContextEvent> iterator() {
				return new IteratorImpl(deserialize, query, params);
			}
		};
	}

	@Override
	public ContextEvent[] query(boolean deserialize, String query, Object... params) {
		return IteratorUtil.toArray(new IteratorImpl(deserialize, query, params));
	}

	@Override
	public Iterable<ContextEvent> loadLazy(final boolean deserialize, final List<Long> contextIds) {
		try {
			return queryLazy(deserialize, dialect.loadContextEvents(contextIds.size()), contextIds.toArray());
		} catch (UnsupportedOperationException e) {
			return new Iterable<ContextEvent>() {
				@Override
				public Iterator<ContextEvent> iterator() {
					List<Iterator<ContextEvent>> loaded = new ArrayList<Iterator<ContextEvent>>();
					for (Long contextId : contextIds)
						loaded.add(new IteratorImpl(deserialize, dialect.loadContextEvent(), contextId));
					return new IteratorUtil.ChainedIterator<ContextEvent>(loaded);
				}
			};
		}
	}

	@Override
	public ContextEvent[] load(boolean deserialize, List<Long> contextIds) {
		try {
			return query(deserialize, dialect.loadContextEvents(contextIds.size()), contextIds.toArray());
		} catch (UnsupportedOperationException e) {
			List<ContextEvent> loaded = new ArrayList<ContextEvent>();
			for (Long contextId : contextIds) {
				ContextEvent loadedEvent = load(deserialize, contextId);
				if (loadedEvent != null)
					loaded.add(loadedEvent);
			}
			return loaded.toArray(new ContextEvent[loaded.size()]);
		}
	}

	@Override
	public ContextEvent load(boolean deserialize, Long contextId) {
		ContextEvent[] cev = query(deserialize, dialect.loadContextEvent(), contextId);
		if (cev.length >= 1)
			return cev[0];
		return null;
	}

	private PreparedStatement parametrizedStatement(String query, Object... params) throws SQLException {
		PreparedStatement stmt = conn.prepareStatement(query);
		for (int i = 0; i < params.length; i++)
			stmt.setObject(i + 1, params[i]);
		return stmt;
	}

	private static Map<String, Integer> columnIndexesForContextEvents(ResultSet rs) throws SQLException {
		HashMap<String, Integer> cols = new HashMap<String, Integer>();
		for (String col : COLS)
			cols.put(col, rs.findColumn(col));
		return cols;
	}

	private void queryLastID() throws SQLException {
		Statement stmt = null;
		try {
			stmt = conn.createStatement();
			stmt.execute(dialect.queryLastID());
			ResultSet rs = stmt.getResultSet();
			if (rs.next())
				lastID = rs.getLong(1);
		} finally {
			if (stmt != null)
				stmt.close();
		}
	}

	@Override
	public Long retrieveLastID() {
		return lastID;
	}

	private class IteratorImpl extends RowIterator {

		private PreparedStatement stmt;
		private ResultSet rows;
		private Map<String, Integer> cols;

		protected IteratorImpl(boolean deserialize, String query, Object... params) {
			super(JDBCStorage.this.log, deserialize, query, params);
		}

		@Override
		protected void initialize() throws IteratorException {
			try {
				stmt = parametrizedStatement(query, params);
				rows = stmt.executeQuery();
				cols = columnIndexesForContextEvents(rows);
			} catch (SQLException e) {
				throw new IteratorException(e);
			}
		}

		@Override
		protected boolean moveNext() throws IteratorException {
			try {
				return rows.next();
			} catch (SQLException e) {
				throw new IteratorException(e);
			}
		}

		@Override
		protected ContextEvent loadContextEvent() throws IteratorException {
			try {
				Long id = rows.getLong(cols.get(COL_ID));
				String originID = rows.getString(cols.get(COL_ORIGINID));
				String ctxstoreID = rows.getString(cols.get(COL_CTXSTORE_ID));
				String type = rows.getString(cols.get(COL_TYPE));
				Date timestamp = new Date(rows.getTimestamp(cols.get(COL_TIMESTAMP)).getTime());
				String[] sdata = new String[] { rows.getString(cols.get(COL_DATA_CLASS)), rows.getString(cols.get(COL_DATA)) };
				Object data = null;
				// there *is* payload data
				if (sdata[0] != null)
					data = (deserialize) ? serializer.deserialize(sdata) : sdata;
				return new ContextEvent(id, data, originID, timestamp, type, ctxstoreID);
			} catch (SQLException e) {
				throw new IteratorException(e);
			}
		}

		@Override
		protected void close() {
			try {
				if (stmt != null)
					stmt.close();
			} catch (SQLException ignore) {
			}
		}

		@Override
		protected void finalize() throws Throwable {
			close();
			super.finalize();
		}
	}
}